import os
import torch
import random
import argparse
import datetime
import math as ma
import numpy as np
import torch.nn as nn
from torch.utils.data import Dataset
from Encode2Decode import Encode2Decode
from torch.utils.data import DataLoader
from torch.optim.lr_scheduler import StepLR
from save_ionex.save_ionex import create_file

# =================================================== config parameters ================================================
model_save_path = './'
frames = 24
rows = 71
cols = 73
device = torch.device("cuda:0")

class ION_Dataset(Dataset):

    def __init__(self, Data, Label):
        self.data = Data
        self.label = Label

    def __getitem__(self, index):
        data = torch.tensor(self.data[index, :, :])
        label = torch.tensor(self.label[index, :, :])
        return data, label

    def __len__(self):
        return len(self.data)

def data_provider(data_paths, train_year, test_year, img_width, args):

    # Training data
    train_npz_files = [data_paths + 'sequence_array_{}.npz'.format(str(x)) for x in range(train_year[0], train_year[1]+1)]
    train_data = load_npz(train_npz_files)
    train_data_number_days = len(train_data) // frames
    train_data = train_data.reshape(train_data_number_days, frames, rows, cols, 1)
    train_data = np.concatenate((train_data, train_data[:, :, :, cols-(img_width[1]-cols):cols, :]), axis=3)
    train_data = np.concatenate((train_data, train_data[:, :, rows-(img_width[1]-rows):rows, :, :]), axis=2)
    max_value = np.max(train_data)
    min_value = np.min(train_data)

    # Test data
    test_npz_files = [data_paths + 'sequence_array_{}.npz'.format(str(x)) for x in range(test_year[0], test_year[1] + 1)]
    test_data = load_npz(test_npz_files)
    test_data_number_days = len(test_data) // frames
    test_data = test_data.reshape(test_data_number_days, frames, rows, cols, 1)
    test_data = np.concatenate((test_data, test_data[:, :, :, cols - (img_width[1] - cols):cols, :]), axis=3)
    test_data = np.concatenate((test_data, test_data[:, :, rows - (img_width[1] - rows):rows, :, :]), axis=2)
    last_train_data = train_data[len(train_data)-args.total_length+1:len(train_data), :, :, :]
    test_data = np.append(last_train_data, test_data, axis=0)

    train_data = Min_Max_Normalization(train_data, max_value, min_value)
    test_data = Min_Max_Normalization(test_data, max_value, min_value)

    train_x, train_y = get_tec_map_x_y(train_data, args.left_length, 1, args.right_length)
    test_x, test_y = get_tec_map_x_y(test_data, args.left_length, 1, args.right_length)

    ion_train_data_set = ION_Dataset(train_x, train_y)
    train_loader = DataLoader(dataset=ion_train_data_set, batch_size=args.batch_size, shuffle=True, pin_memory=True, num_workers=0)
    ion_test_data_set = ION_Dataset(test_x, test_y)
    test_loader = DataLoader(dataset=ion_test_data_set, batch_size=args.batch_size, shuffle=False, pin_memory=True, num_workers=0)

    return train_loader, test_loader, max_value, min_value

def load_npz(names):

    sequence = np.load(names[0])['sequence_array']
    for i in range(1, len(names)):
        sequence = np.concatenate((sequence, np.load(names[i])['sequence_array']), axis=0)
    print(sequence.shape)
    return sequence

def get_tec_map_x_y(all_data, num_periods_input, gap_hour, num_periods_output):

    start = 0
    next = 0
    end = len(all_data)
    x_batches = []
    y_batches = []

    limit = num_periods_input + num_periods_output
    while start + (limit) <= end:
        next = start + num_periods_input
        x_batches.append(all_data[start:start+num_periods_input, :, :, :, :])
        y_batches.append(all_data[next:next+num_periods_output, :, :, :, :])
        start = start + gap_hour

    x_batches = np.asarray(x_batches, dtype=np.float32)
    x_batches = x_batches.reshape(-1, num_periods_input*24, 75, 75, 1)

    y_batches = np.asarray(y_batches, dtype=np.float32)
    y_batches = y_batches.reshape(-1, 24, 75, 75, 1)

    return x_batches, y_batches

def Min_Max_Normalization(data, max, min):
    return (data - min) / (max - min)

def Min_Max_Unnormalization(data, max, min):
    return data * (max - min) + min

def train_loop(train_loader, model, optimizer, args, itr, max_value, min_value, loss_function):

    model.train()

    if args.use_amp:
        scaler = torch.cuda.amp.GradScaler()

    TRAIN_loss = []
    if args.use_amp:
        for batch, [x, y] in enumerate(train_loader):
            with torch.cuda.amp.autocast():
                ims = x
                output_ims = y
                ims = ims.permute(0, 4, 1, 2, 3).contiguous()
                output_ims = output_ims.permute(0, 4, 1, 2, 3).contiguous()
                ims, output_ims = ims.half(), output_ims.half()
                ims, output_ims = ims.to(device), output_ims.to(device)

                y_pred, loss, loss_l1 = model(ims, output_ims)
                # caluate the loss
                y_ture = y[:, :, 0:71, 0:73, :].to(device)
                label_value = Min_Max_Unnormalization(y_ture, max_value, min_value)
                y_pred_ture = y_pred.permute(0, 2, 3, 4, 1).contiguous()[:, :, 0:71, 0:73, :]
                y_pred_value = Min_Max_Unnormalization(y_pred_ture, max_value, min_value)
                true_train_loss = loss_function(label_value, y_pred_value)
                TRAIN_loss.append(true_train_loss.item())

                total_loss = loss_l1
            # Back propagation
            optimizer.zero_grad()
            scaler.scale(total_loss.mean()).backward()
            scaler.step(optimizer)
            scaler.update()
    else:
        for batch, [x, y] in enumerate(train_loader):
            ims = x
            output_ims = y
            ims = ims.permute(0, 4, 1, 2, 3).contiguous()
            output_ims = output_ims.permute(0, 4, 1, 2, 3).contiguous()
            ims, output_ims = ims.to(device), output_ims.to(device)

            y_pred, loss, loss_l1 = model(ims, output_ims)
            # caluate the loss
            y_ture = y[:, :, 0:71, 0:73, :].to(device)
            label_value = Min_Max_Unnormalization(y_ture, max_value, min_value)
            y_pred_ture = y_pred.permute(0, 2, 3, 4, 1).contiguous()[:, :, 0:71, 0:73, :]
            y_pred_value = Min_Max_Unnormalization(y_pred_ture, max_value, min_value)
            true_train_loss = loss_function(label_value, y_pred_value)
            TRAIN_loss.append(true_train_loss.item())

            total_loss = loss_l1
            # Back propagation
            optimizer.zero_grad()
            scaler.scale(total_loss).backward()
            scaler.step(optimizer)
            scaler.update()

    print('-------------------------------', 'itr: ' + str(itr), 'training loss: ' + str(ma.sqrt(np.mean(TRAIN_loss))), '-------------------------------')
    return ma.sqrt(np.mean(TRAIN_loss))

def test_loop(test_loader, model, itr, max_value, min_value, loss_function):

    model.eval()

    result_data = []
    test_loss = []
    bias_loss = []

    with torch.no_grad():
        for test_x, test_y in test_loader:
            ims = test_x
            output_ims = test_y
            ims = ims.permute(0, 4, 1, 2, 3).contiguous()
            output_ims = output_ims.permute(0, 4, 1, 2, 3).contiguous()
            ims, output_ims = ims.to(device), output_ims.to(device)
            y_pred, loss, loss_l1 = model(ims, output_ims)
            y_ture = test_y[:, :, 0:71, 0:73, :].to(device)
            label_value = Min_Max_Unnormalization(y_ture, max_value, min_value)
            y_pred_ture = y_pred.permute(0, 2, 3, 4, 1).contiguous()[:, :, 0:71, 0:73, :]
            y_pred_value = Min_Max_Unnormalization(y_pred_ture, max_value, min_value)
            # total loss
            true_test_loss = loss_function(label_value, y_pred_value)
            test_loss.append(true_test_loss.item())
            # every grid loss
            true_bias_loss = label_value - y_pred_value
            bias_loss.append(true_bias_loss.cpu().numpy())
            # sess_result
            sess_result = y_pred_value.cpu().numpy().tolist()
            for d in range(0, len(sess_result), 1):
                day_prediction_data = np.array(sess_result[d])
                result_data.append(day_prediction_data)

        model_bias = bias_loss[0]
        for b in range(1, len(bias_loss), 1):
            model_bias = np.concatenate((model_bias, bias_loss[b]), axis=0)

        model_MSE = model_bias * model_bias

    return result_data, np.mean(model_bias), np.sqrt(np.mean(model_MSE))

def count_days(year, month, day):

    d = datetime.date(year, month, day)
    d0 = datetime.date(year, 1, 1)
    return (d - d0).days

def seed_everything(seed=3407):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
# ======================================================================================================================
if __name__ == '__main__':

    SEED = 3407
    seed_everything(SEED)
    # ------------------------------------------------------------------------------------------------------------------
    parser = argparse.ArgumentParser(description='SAM-ConvLSTM of VTEC')

    # data
    parser.add_argument('--total_length', type=int, default=3 + 1)
    parser.add_argument('--left_length', type=int, default=3)
    parser.add_argument('--right_length', type=int, default=1)
    parser.add_argument('--input_total_length', type=int, default=3)
    parser.add_argument('--img_width', type=int, default=75)
    parser.add_argument('--img_channel', type=int, default=1)

    # model
    parser.add_argument('--model_name', type=str, default='SAM-ConvLSTM')
    parser.add_argument('--num_hidden', type=int, default=128)
    parser.add_argument('--attention_num_hidden', type=int, default=64)
    parser.add_argument('--layers_num', type=int, default=3)
    parser.add_argument('--filter_size', type=int, default=3)
    parser.add_argument('--stride', type=int, default=1)
    parser.add_argument('--patch_size', type=int, default=5)
    parser.add_argument('--layer_norm', type=int, default=1)
    parser.add_argument('--activate_function', type=str, default='tanh')
    parser.add_argument('--frame_size', default=(19, 19))
    parser.add_argument('--use_amp', type=int, default=1)

    # optimization
    parser.add_argument('--lr', type=float, default=0.00001)
    parser.add_argument('--batch_size', type=int, default=16)
    parser.add_argument('--max_iterations', type=int, default=252)
    parser.add_argument('--n_gpu', type=int, default=0)

    args = parser.parse_args()
    # ------------------------------------------------------------------------------------------------------------------
    train_set, test_set, max_value, min_value = data_provider('/home/ps/codg_data/', [2008, 2014], [2015, 2015], [75, 75], args)
    # ------------------------------------------------------------------------------------------------------------------
    model = Encode2Decode(args.img_channel, args.num_hidden, args.attention_num_hidden, configs=args, kernel_size=(3, 3), img_size=args.frame_size)
    if torch.cuda.device_count() > 1:
        model = nn.DataParallel(model, device_ids=[0, 1])
        model.to(device)
    # ------------------------------------------------------------------------------------------------------------------
    optimizer = torch.optim.Adam(model.parameters(), lr=args.lr, weight_decay=1e-4, eps=1e-08)
    scheduler = StepLR(optimizer, step_size=100, gamma=0.5)
    # ------------------------------------------------------------------------------------------------------------------
    loss_function = nn.MSELoss()
    loss_function = loss_function.to(device)
    # ------------------------------------------------------------------------------------------------------------------
    for itr in range(1, args.max_iterations + 1):

        train_loss = train_loop(train_set, model, optimizer, args, itr, max_value, min_value, loss_function)
        
        scheduler.step()

        result, model_bias, model_mse = test_loop(test_set, model, itr, max_value, min_value, loss_function)
    # ----------------------------------------------- get the prediction files -----------------------------------------
    result, model_bias, model_mse = test_loop(test_set, model, args.max_iterations, max_value, min_value, loss_function)

    fmt_name = "sacl{:03}0.{}i"
    date_obs = datetime.date(2015, 1, 1)
    for i in range(0, len(result)):
        tec_data = result[i][:, :, :, 0]
        cur_date = date_obs + datetime.timedelta(days=i)
        new_index = count_days(cur_date.year, cur_date.month, cur_date.day)
        name = fmt_name.format(new_index + 1, str(cur_date.year)[-2:])
        create_file(os.path.join("./prediction/2015/", name), cur_date.year, cur_date.month, cur_date.day, tec_data)
        print("已生成预测得到的文件：" + "./prediction/2015/" + name)
