import torch
import random
import math as ma
import torch.nn as nn
from SAConvLSTM import SAConvLSTMCell


class Encode2Decode(nn.Module):
   
    # self-attention convlstm for spatiotemporal prediction model
    def __init__(self, input_dim, hidden_dim, attn_hidden_dim, kernel_size, configs, img_size=(16, 16), num_layers=4,
                 batch_first=True,
                 bias=True,
                 ):
        super(Encode2Decode, self).__init__()
        self.img_size = img_size
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.attn_hidden_dim = attn_hidden_dim
        self.kernel_size = kernel_size
        self.configs = configs
        self.num_layers = num_layers
        self.batch_first = batch_first
        self.bias = bias
        self.MSE_criterion = nn.MSELoss()
        self.MAE_criterion = nn.L1Loss()
        
        self.img_encode = nn.Sequential(
            nn.Conv2d(input_dim, hidden_dim, (1, 1), (1, 1)),
            nn.LeakyReLU(0.1),
            nn.Conv2d(hidden_dim, hidden_dim, (3, 3), (2, 2), (1, 1)),
            nn.LeakyReLU(0.1),
            nn.Conv2d(hidden_dim, hidden_dim, (3, 3), (2, 2), (1, 1)),
            nn.LeakyReLU(0.1)
        )
        
        self.img_decode = nn.Sequential(
            nn.ConvTranspose2d(hidden_dim, hidden_dim, (3, 3), (2, 2), (1, 1), output_padding=(1, 1)),
            nn.LeakyReLU(0.1),
            nn.ConvTranspose2d(hidden_dim, hidden_dim, (3, 3), (2, 2), (1, 1), output_padding=(1, 1)),
            nn.LeakyReLU(0.1),
            nn.Conv2d(hidden_dim, input_dim, (1, 1), (1, 1)),
        )
        cell_list, bns = [], []
        for i in range(0, self.num_layers):
            cur_input_dim = self.hidden_dim

            # the sam-convlstm layer
            cell_list.append(SAConvLSTMCell(input_dim=cur_input_dim,
                                            hidden_dim=self.hidden_dim,
                                            kernel_size=self.kernel_size,
                                            att_hidden_dim=self.attn_hidden_dim,
                                            bias=self.bias,
                                            ))
            # Use layer norm
            bns.append(nn.LayerNorm(normalized_shape=[hidden_dim, *self.img_size]))

        self.cell_list = nn.ModuleList(cell_list)
        self.bns = nn.ModuleList(bns)

        # last layer using Linear to get output
        self.decoder_predict = nn.Conv2d(in_channels=hidden_dim,
                                         out_channels=input_dim,
                                         kernel_size=(1, 1),
                                         padding=(0, 0))

    def forward(self, frames_tensor, label_tensor, hidden_state=None):

        frames = frames_tensor.permute(0, 2, 1, 3, 4).contiguous()
        b, l, c, h, w = frames.shape

        if hidden_state is None:
            hidden_state = self._init_hidden(batch_size=b, image_size=(ma.ceil(h / 4), ma.ceil(w / 4)), device=frames.device)
        next_frames = []

        for t in range((self.configs.left_length-self.configs.input_total_length) * 24, (self.configs.total_length) * 24, 1):
            # input data
            if t < (self.configs.left_length + 0) * 24:
                net = frames[:, t]
            else:
                net = frames[:, t-24]
            net = net.float()

            x = self.img_encode(net)

            for i, cell in enumerate(self.cell_list):
                h_next, c_next, m_next = cell(x, hidden_state[i])
                out, hidden_state[i] = h_next, (h_next, c_next, m_next)
                out = self.bns[i](out)
            out = self.img_decode(out)
            if t >= (self.configs.left_length + 0) * 24:
                next_frames.append(out[:, :, 0:71, 0:73])

        next_frames = torch.stack(next_frames, dim=0).permute(1, 2, 0, 3, 4).contiguous()
        loss_l1 = self.MSE_criterion(next_frames, label_tensor[:, :, :, 0:71, 0:73])
        loss_l2 = self.MAE_criterion(next_frames, label_tensor[:, :, :, 0:71, 0:73])
        loss = loss_l1 + loss_l2

        return next_frames, loss, loss_l1

    def _init_hidden(self, batch_size, image_size, device):
        init_states = []
        for i in range(self.num_layers):
            init_states.append(self.cell_list[i].init_hidden(batch_size, image_size, device))
        return init_states


def main():
    model = Encode2Decode(1, 16, 16, (3, 3))
    img_size = 1, 64, 64
    batch_size, seq_len, horizon = 2, 10, 10
    x = torch.rand(batch_size, seq_len, *img_size)
    y = torch.rand(batch_size, horizon, *img_size)
    y_hat = model(x, y)
    print(y_hat.shape)


if __name__ == '__main__':
    main()
